/********************************************************************
Code by Julien Roh. 
Thanks Niklas Buratti and Cecilia Vecchione for the precious help.
Inspired by the original "Conway's Game of Life" (1970)
2D KSSB 2023
********************************************************************/

// Set the size of the grid and each cell
int gridSize = 100;
int cellSize = 15; // set the size of the pixels
int currentFrameRate = 10;// int for the framerate

int[][] grid = new int[gridSize][gridSize]; //create a 2D array to represent the grid and a boolean to track if the simulation is running 
boolean isPlaying = false;

// Add a new enum for the game state
enum Screen {
  MENU, GAME, SETTINGS
}

// Add a new enum for color schemes
enum ColorScheme {
  DEFAULT, DARK_GREEN, PASTEL, ORANGE_PURPLE
}

// Define color scheme data structure
class GameColorScheme {
  color backgroundColor;
  color strokeColor;
  color cellColor;

  GameColorScheme(color backgroundColor, color strokeColor, color cellColor) {
    this.backgroundColor = backgroundColor;
    this.strokeColor = strokeColor;
    this.cellColor = cellColor;
  }
}

Screen currentScreen = Screen.MENU; // Set initial game state to MENU
ColorScheme currentColorSchemeType = ColorScheme.DEFAULT; // Set initial color scheme to DEFAULT
GameColorScheme currentColorScheme;

// Define color schemes
GameColorScheme defaultColorScheme = new GameColorScheme(color(255), color(200), color(0));
GameColorScheme darkGreenColorScheme = new GameColorScheme(color(0, 50, 0), color(0), color(0, 255, 0));
GameColorScheme pastelColorScheme = new GameColorScheme(color(173, 216, 230), color(255), color(255, 182, 193));
GameColorScheme orangePurpleColorScheme = new GameColorScheme(color(128, 0, 128), color(0), color(255, 165, 0));

GameColorScheme[] colorSchemes = new GameColorScheme[]{
  defaultColorScheme,
  darkGreenColorScheme,
  pastelColorScheme,
  orangePurpleColorScheme
};

void setup() {
  size(800, 600); // set the size of the program
  surface.setResizable(true);
  frameRate(10); // numbers of frames per seconds
  noStroke(); // disable the strokes for the shapes
  initializeGrid(); // initialize the grid with all cells set to 0 (dead) -> is also used when pressing "r" to reset the grid
  currentColorScheme = colorSchemes[currentColorSchemeType.ordinal()]; // Set the initial color scheme
}

void draw() {
  switch (currentScreen) {
    case MENU:
      renderMenu();
      break;
    case SETTINGS:
      renderSettings();
      break;
    case GAME:
      renderGrid(); // render the grid
      // If the simulation is running, update the grid
      if (isPlaying) {
        updateGrid();
      }
      break;
  }
}

// Initialize the grid with all cells set to 0 ( so that the player can place them)
void initializeGrid() {
  for (int i = 0; i < gridSize; i++) {
    for (int j = 0; j < gridSize; j++) {
      grid[i][j] = 0;
    }
  }
}

// (this section) update the grid based on the rules of the original game: Conway's Game of Life
void updateGrid() {
  int[][] newGrid = new int[gridSize][gridSize];
  // loop through each cell in the grid
  for (int i = 0; i < gridSize; i++) {
    for (int j = 0; j < gridSize; j++) {
      // get the current state of the cell and the number of living neighbors
      int state = grid[i][j];
      int neighbors = countNeighbors(i, j);

      if (state == 1 && (neighbors < 2 || neighbors > 3)) {
        newGrid[i][j] = 0; // cell dies due to over/underpopulation
      } else if (state == 0 && neighbors == 3) {
        newGrid[i][j] = 1; // a new cell is born
      } else {
        newGrid[i][j] = state; // the cell remains in its current state
      }
    }
  }

  grid = newGrid; // Update the grid with the new state
}

// Render menu screen
void renderMenu() {
  background(currentColorScheme.backgroundColor);
  textAlign(CENTER, CENTER);
  textSize(48);
  fill(currentColorScheme.cellColor);
  text("Conway's Game of Life", width / 2, height / 2 - 60);
  textSize(24);
  text("Press P to Play", width / 2, height / 2);
  text("Press S for Settings", width / 2, height / 2 + 30);
  text("Press Q to Quit", width / 2, height / 2 + 60);
}
// Render settings screen
void renderSettings() {
  background(currentColorScheme.backgroundColor);
  textAlign(CENTER, CENTER);
  textSize(24);
  fill(currentColorScheme.cellColor);
  text("Settings", width / 2, height / 2 - 90);
  text("1. Default (Black & White)", width / 2, height / 2 - 60);
  text("2. Dark Theme (Green Pixels)", width / 2, height / 2 - 30);
  text("3. Pastel Theme (Pink Pixels)", width / 2, height / 2);
  text("4. Vibrant Theme (Orange Pixels)", width / 2, height / 2 + 30);
  text("Frame Rate: " + currentFrameRate + " (Press +/- to change)", width / 2, height / 2 + 60);
  text("Press M to return to the menu", width / 2, height / 2 + 90);
  
  textAlign(LEFT, BOTTOM);
  textSize(16);
  text("Keep in mind that setting the frame rate too high may cause performance issues on slower devices. ", 10, height - 30);
  textAlign(LEFT, BOTTOM);
  textSize(16);
  text("That's the same with the window size ;) ", 10, height - 10);
}

void renderGrid() { // Render the grid as a set of rectangles
  background(currentColorScheme.backgroundColor); // Set the background color based on the current color scheme

  // Loop through each cell in the grid
  for (int i = 0; i < gridSize; i++) {
    for (int j = 0; j < gridSize; j++) {
      // Set the fill color based on the state of the cell and the current color scheme
      fill(grid[i][j] == 1 ? currentColorScheme.cellColor : currentColorScheme.backgroundColor);

      // Draw a rectangle for the cell
      rect(i * cellSize, j * cellSize, cellSize, cellSize);

      // Draw gridlines to separate the cells based on the current color scheme
      stroke(currentColorScheme.strokeColor);
      line(i * cellSize, 0, i * cellSize, gridSize * cellSize);
      line(0, j * cellSize, gridSize * cellSize, j * cellSize);
    }
  }
}

int countNeighbors(int x, int y) {
  int count = 0;
  for (int i = -1; i <= 1; i++) {
    for (int j = -1; j <= 1; j++) {
      if (i == 0 && j == 0) continue;
      int nx = (x + i + gridSize) % gridSize;
      int ny = (y + j + gridSize) % gridSize;
      count += grid[nx][ny];
    }
  }
  return count;
}
void mousePressed() {
  if (mouseButton == LEFT) {
    if (currentScreen == Screen.GAME) {
      int i = mouseX / cellSize;
      int j = mouseY / cellSize;
      grid[i][j] = (grid[i][j] + 1) % 2;
    } else if (currentScreen == Screen.MENU) {
      if (overPlayButton()) {
        currentScreen = Screen.GAME;
      } else if (overSettingsButton()) {
        currentScreen = Screen.SETTINGS;
      } else if (overQuitButton()) {
        exit();
      }
    }
  }
}

//Keys here are to navigate in the menu or interact with the game
void keyPressed() {
  if (currentScreen == Screen.GAME) {
    // When the "r" key is pressed, it resets the grid
    if (key == 'r' || key == 'R') {
      initializeGrid();
    }
    if (key == 'm' || key == 'M') {
      currentScreen = Screen.MENU;
    }
    // The spacebar will start the game (or stop it if the game is launched)
    if (key == ' ') {
      isPlaying = !isPlaying;
    }
  } else if (currentScreen == Screen.SETTINGS) {
    if (key >= '1' && key <= '4') {
      int index = key - '1';
      currentColorScheme = colorSchemes[index];
    }
    if (key == '+' || key == '=') {
      currentFrameRate = min(currentFrameRate + 5, 60);
      frameRate(currentFrameRate);
    }
    if (key == '-') {
      currentFrameRate = max(currentFrameRate - 5, 1);
      frameRate(currentFrameRate);
    }
    if (key == 'm' || key == 'M') {
      currentScreen = Screen.MENU;
    }
  } else if (currentScreen == Screen.MENU) {
    if (key == 'P' || key == 'p') {
        currentScreen = Screen.GAME;
    } else if (key == 'S' || key == 's') {
        currentScreen = Screen.SETTINGS;
    } else if (key == 'Q' || key == 'q') {
        exit();
    }
  }
}


boolean overPlayButton() {
  return mouseX > (width / 2 - 60) && mouseX < (width / 2 + 60) && mouseY > (height / 2 - 60) && mouseY < (height / 2 - 30);
}

boolean overSettingsButton() {
  return mouseX > (width / 2 - 60) && mouseX < (width / 2 + 60) && mouseY > (height / 2 - 20) && mouseY < (height / 2 + 10);
}
boolean overQuitButton() {
  return mouseX > (width / 2 - 60) && mouseX < (width / 2 + 60) && mouseY > (height / 2 + 20) && mouseY < (height / 2 + 50);
}

void applyColorScheme() {//(ColorScheme currentColorScheme) {
  background(currentColorScheme.backgroundColor);
  stroke(currentColorScheme.strokeColor);
}
